#
# Christmas Logic 
# Copyright (C) Jack Whitham 2009
#

import wx, math, collections, array, sys, random, time
from wx.lib.newevent import NewEvent

import font, guide
from constants import *
from disk import RO_DIR, RW_DIR

(AckButtonEvent, EVT_ACK_BUTTON) = NewEvent()

RB_CLOCK_EVENT = (ST_CLOCK, (128, 128, 128)) 
SIZE_LUT = {
        ST_CLOCK :      SEQ_ONE_HEIGHT + 5,
        ST_ONE :        SEQ_ONE_HEIGHT - 2,
        ST_UNDEFINED :  SEQ_ONE_HEIGHT - 2,
        ST_ZERO :       SEQ_ZERO_HEIGHT,
        }

class Tree(wx.Window):
    def __init__(self, parent, *args1, **args2):
        wx.Window.__init__(self, parent, *args1, **args2)
        self.SetBackgroundColour((0, 0, 0))
        self.Bind(wx.EVT_PAINT, self.On_Paint)

        self.network = None
        self.blank = True

        size = TREE_SIZE
        self.SetMinSize(size)
        self.SetMaxSize(size)
        self.SetSize(size)
        self.SetSizeHints(size[ 0 ], size[ 1 ])

        self.cso_image = None

    def Unset_Network(self):
        self.network = None
        
    def Set_Network(self, n):
        self.network = n
        self.blank = True
        self.Refresh(False)

    def Set_Image(self, cso_image):
        self.cso_image = cso_image
        self.blank = True
        self.Refresh(False)

    def On_Paint(self, event):
        if OVERPAINT_DEBUG:
            print '%1.2f Paint Tree' % (time.time() % 1000)

        dc = wx.PaintDC(self)
        if self.cso_image == None:
            return
        if self.network == None:
            return
        if self.blank:
            dc.SetPen(wx.Pen((0, 0, 0)))
            dc.SetBrush(wx.Brush((0, 0, 0)))
            (w, h) = self.GetSize()
            dc.DrawRectangle(0, 0, w, h)
            self.blank = False

        dc.DrawBitmap(self.cso_image.Get_Base(), 0, 0)

        for (i, s) in enumerate(self.network.Get_String_Data()):
            if s.state != ST_UNDEFINED:
                dc.DrawBitmap(
                    self.cso_image.Get_Overlay(i, s.state), 
                    0, 0, True)

class Logical(wx.ScrolledWindow):
    def __init__(self, parent, *args1, **args2):
        args2[ 'style' ] = (args2.get('style', 0) | 
                    wx.VSCROLL | wx.HSCROLL | wx.ALWAYS_SHOW_SB)
        wx.ScrolledWindow.__init__(self, parent, *args1, **args2)
        self.SetBackgroundColour((255, 255, 255))
        self.Bind(wx.EVT_PAINT, self.On_Paint)
        self.Bind(wx.EVT_LEFT_DOWN, self.On_Mouse_Down)
        self.Bind(wx.EVT_LEFT_UP, self.On_Mouse_Up)
        self.Bind(wx.EVT_RIGHT_DOWN, self.On_Mouse_Down)
        self.Bind(wx.EVT_RIGHT_UP, self.On_Mouse_Up)
        self.Bind(wx.EVT_MOTION, self.On_Mouse_Motion)
        self.Bind(wx.EVT_SIZE, self.On_Size_Event)
        self.Bind(wx.EVT_TIMER, self.On_Timer)

        self.ack_button = wx.Button(self, -1, "pXYZZYq")
        self.show_ack_button = False
        self.ack_button.Hide()
        self.ack_button.Bind(wx.EVT_BUTTON, self.On_Ack_Button)
        self.ack_button.SetFont(wx.Font(16, wx.ROMAN, 
                    wx.NORMAL, wx.NORMAL, 0, ""))

        self.ensure_visible = None
        self.refresh_value = 0
        self.refresh_rects = []

        self.timer = wx.Timer(self, wx.NewId())
        self.timer.Start(SCROLL_TIMER_INTERVAL, False)

        self.network = None
        self.scroll_bar_setting = dict()
        self.origin_x = self.origin_y = 0
        self.next_origin_x = self.next_origin_y = 0

        #self.SetMinSize(size)
        #self.SetSizeHints(size[ 0 ], size[ 1 ])
        self.guide = guide.Guide()

        self.Change_Scrollbar(mx1 = 0, my1 = 0, x2 = 1, y2 = 1)
        self.fail = False

    def On_Ack_Button(self, event):
        self.ack_counter += 1
        wx.PostEvent(self, AckButtonEvent())

    def Update_Label(self):
        if self.network == None:
            return 
    
        rc = self.network.Update_Label(self.ack_counter)
        loc = self.guide.Setup(self, **rc)
        requires_ack = rc.get("requires_ack", False)
        label = rc.get("ack_text", "Next...")
        if requires_ack != self.show_ack_button:
            self.show_ack_button = requires_ack
            if requires_ack:
                self.ack_button.Show()
            else:
                self.ack_button.Hide()
           
        if requires_ack:
            assert loc != None
            (x, y, w) = loc
            (w1, h) = self.ack_button.GetSize()
            self.ack_button.SetPosition(self.Unoffset((x, y)))
            self.ack_button.SetSize((w, h))
            if label != self.ack_button.GetLabel():
                self.ack_button.SetLabel(label)
            
    def Update_Area(self, (x, y), (w, h)):
        #self.refresh_rects.append((x, y, w, h))
        (x, y) = self.Unoffset((x, y))
        wx.Window.RefreshRect(self, (x - 1, y - 1, w + 2, h + 2))

    def Alert_Popup(self, error):
        import logic
        assert isinstance(error, logic.Network_Error)
        dlg = wx.MessageDialog(self,
                message = error.Get_Message(),
                caption = APP_NAME + " - Not Possible",
                style = wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def On_Size_Event(self, event):
        self.Refresh()

    def Change_Scrollbar(self, **args):
        (w, h) = self.GetSize()
        vsize = []
        for (key1, key2) in [ ("mx1", "x2"), ("my1", "y2") ]:
            x1 = - args[ key1 ] 
            x2 = args[ key2 ]
            vsize.append(x2 - x1)

        self.SetVirtualSize(tuple(vsize))
        self.SetScrollRate(1, 1)
        self.next_origin_x = - args[ "mx1" ]
        self.next_origin_y = - args[ "my1" ]
        self.scroll_bar_setting = args
        self.Refresh()

    def Ensure_Visible(self, (x, y)):
        (w, h) = self.GetSize()
        (vx, vy) = self.Get_View_XY()
        complete = True
        out = []
        for (view_base, objx, size) in [ (vx, x, w), (vy, y, h) ]:
            dx = 0
            view_limit = view_base + size
            if objx <= view_base:
                # Scroll left
                dx = - (view_base - objx)
            elif objx >= view_limit:
                # Scroll right
                dx = objx - view_base

            dx2 = min(SCROLL_SPEED, max(- SCROLL_SPEED, dx))
            if dx2 != dx:
                complete = False

            out.append(dx2)
           
        self.MGC_Scroll(tuple(out))

        self.ensure_visible = None
        if not complete:
            self.ensure_visible = (x, y)

    def On_Timer(self, event):
        if self.ensure_visible != None:
            self.Ensure_Visible(self.ensure_visible)

    def MGC_Scroll(self, (dx, dy)):
        # MGC = maps.famoussearchengine.com
        (x, y) = self.GetViewStart()
        self.Scroll(dx + x, dy + y)

    def Offset(self, (x, y)):
        (vx, vy) = self.Get_View_XY()
        return (x + vx, y + vy)

    def Unoffset(self, (x, y)):
        (vx, vy) = self.Get_View_XY()
        return (x - vx, y - vy)

    def Get_View_XY(self):
        (x, y) = self.GetViewStart()
        x += self.origin_x
        y += self.origin_y
        return (x, y)

    def On_Paint(self, event):
        if self.fail:
            sys.exit(1)

        if OVERPAINT_DEBUG:
            print '%1.2f Paint Logical' % (time.time() % 1000)

        self.origin_x = self.next_origin_x
        self.origin_y = self.next_origin_y
        dc = wx.PaintDC(self)
        (x, y) = self.Get_View_XY()
        dc.SetDeviceOrigin(-x, -y)

        if self.network != None:
            try:
                self.network.Paint(dc)
            except:
                self.fail = True
                raise

            self.guide.Draw(dc)


    def Unset_Network(self):
        if self.network != None:
            self.network.WX_Disconnect()
            self.network = None
        
    def Set_Network(self, n):
        self.Unset_Network()
        self.ack_counter = 0
        self.network = n
        self.network.WX_Connect(self)
        self.Refresh()
        #print 'Refresh logical (set network)'

    def On_Mouse_Down(self, event):
        if self.network == None:
            return
        if self.fail:
            sys.exit(1)
        try:
            self.network.Mouse_Down(
                    self.Offset(event.GetPositionTuple()))
        except:
            self.fail = True
            raise

    def On_Mouse_Up(self, event):
        if self.network == None:
            return
        if self.fail:
            sys.exit(1)
        try:
            self.network.Mouse_Up(
                    self.Offset(event.GetPositionTuple()))
        except:
            self.fail = True
            raise

    def On_Mouse_Motion(self, event):
        if self.network == None:
            return
        if self.fail:
            sys.exit(1)
        try:
            self.network.Mouse_Motion(
                    self.Offset(event.GetPositionTuple()))
        except:
            self.fail = True
            raise


class Out_Sequence(wx.Window):
    def __init__(self, parent, *args1, **args2):
        wx.Window.__init__(self, parent, *args1, **args2)
        self.SetBackgroundColour((0, 0, 0))
        self.Bind(wx.EVT_PAINT, self.On_Paint)
        self.stretch = STRETCH
        self.network = None
        self.pcache = dict()
        self.pcache[ 0 ] = wx.Pen((0, 0, 0))
        self.PEN_WHITE = wx.Pen((255, 255, 255))
        self.FONT_SMALL = font.Make_Font(wx.FONTFAMILY_TELETYPE, SEQ_ONE_HEIGHT)
        self.ring_buffer = collections.deque()
        self.max_in_ring_buffer = 1
        self.dirty = True
        self.margin = 0
        self.Set_Label_List([])
        self.Set_Seq_Lines(2, TREE_SIZE[ 0 ])

    def Push(self, new_element):
        self.ring_buffer.append(new_element)
        while len(self.ring_buffer) > self.max_in_ring_buffer:
            self.ring_buffer.popleft()
        self.dirty = True
        self.Refresh(False)

    def Set_Label_List(self, ll):
        self.label_list = ll
        self.dirty = True
        self.Refresh(False)

    def Set_Seq_Lines(self, n, w):
        (w, h) = size = (w, (n * SEQ_ONE_HEIGHT) + (SEQ_MARGIN * 2))
        self.SetMinSize(size)
        self.SetMaxSize(size)
        self.SetSize(size)
        self.SetSizeHints(w, h)
        self.max_in_ring_buffer = (
            (w - (SEQ_MARGIN + SEQ_X_MARGIN)) / self.stretch)
        self.bitmap = wx.EmptyBitmap(w, h)
        self.dirty = True
        self.Refresh(False)

    def Unset_Network(self):
        self.network = None
        self.dirty = True
        self.ring_buffer.clear()
        self.Set_Label_List([])
        
    def Set_Network(self, n):
        self.network = n
        self.dirty = True
        self.Refresh(False)

    def On_Paint(self, event):
        if self.dirty:
            mdc = wx.MemoryDC()
            mdc.SelectObject(self.bitmap)
            self.Paint_Graph(mdc)
            mdc.SelectObject(wx.NullBitmap)
            self.dirty = False

        dc = wx.PaintDC(self)
        dc.DrawBitmap(self.bitmap, 0, 0)

    def Paint_Graph(self, dc):
        x = SEQ_X_MARGIN
        y = 0
        (fw, fh) = self.GetSize()

        dc.SetPen(self.pcache[ 0 ])
        dc.SetBrush(wx.Brush((0, 0, 0)))
        dc.DrawRectangle(0, 0, fw, fh)

        if ((len(self.label_list) == 0)
        and (len(self.ring_buffer) == 0)):
            # Skip all drawing
            return

        dc.SetFont(self.FONT_SMALL)
        for name in self.label_list:
            (w, h) = dc.GetTextExtent(name)
            x = max(w + 2, x)

        self.margin = x

        for signal_list in self.ring_buffer:
            for i in xrange(self.stretch):
                #dc.SetPen(self.pcache[ 0 ])
                #dc.DrawLine(x, 0, x, fh)

                y = SEQ_MARGIN
                for (state, colour) in signal_list:
                    c = colour
                    if state == ST_UNDEFINED:
                        c = COLOUR_UNDEFINED

                    if not self.pcache.has_key(c):
                        self.pcache[ c ] = wx.Pen(c)
                        
                    dc.SetPen(self.pcache[ c ])
                    yh = SIZE_LUT[ state ]
                  
                    y += SEQ_ONE_HEIGHT
                    dc.DrawLine(x, y - yh, x, y)

            
                x += 1

        # Blank right-hand edge
        #dc.SetPen(self.pcache[ 0 ])
        #dc.DrawLine(x, 0, x, fh)
        #dc.DrawLine(x + 1, 0, x + 1, fh)

        self.Paint_Thumb(dc, x, y)

        # Text has a shadow
        dc.SetTextForeground((255, 255, 255))
        y = SEQ_MARGIN 
        for name in self.label_list:
            dc.DrawText(name, 1, y)
            y += SEQ_ONE_HEIGHT

    def Paint_Thumb(self, dc, x, y):
        dc.SetPen(self.PEN_WHITE)
        dc.DrawLine(x, SEQ_MARGIN / 2, x, y + (SEQ_MARGIN / 2))
        x += 1
        dc.DrawLine(x, SEQ_MARGIN / 2, x, y + (SEQ_MARGIN / 2))

class Goal_Sequence(Out_Sequence):
    def __init__(self, *args1, **args2):
        Out_Sequence.__init__(self, *args1, **args2)
        self.thumb_pos = -1

    def Update(self):
        self.ring_buffer.clear()
        self.dirty = True

        if self.network == None:
            return 0

        gsl = self.network.Get_Number_of_Goals()
        
        for i in xrange(gsl):
            for j in xrange(CLOCK_CYCLE_LENGTH - 1):
                self.ring_buffer.append([
                    (s.goal_sequence[ i ], s.colour) 
                        for s in self.network.Get_String_Data() ])
            self.ring_buffer.append([ RB_CLOCK_EVENT
                        for s in self.network.Get_String_Data() ])
       
        self.stretch = STRETCH
        if gsl > 6:
            self.stretch = 1

        return len(self.ring_buffer[ 0 ])

    def Paint_Thumb(self, dc, x, y):
        Out_Sequence.Paint_Thumb(self, dc, 
                    self.thumb_pos + self.margin, y)

    def Set_Thumb_Pos(self, tp):
        if tp != self.thumb_pos:
            self.thumb_pos = tp * self.stretch
            self.dirty = True
            self.Refresh(False)

class In_Sequence(Goal_Sequence):
    def __init__(self, *args1, **args2):
        self.dis_input = False
        Goal_Sequence.__init__(self, *args1, **args2)

    def Set_Dis_Input(self, d):
        self.dirty = True
        self.dis_input = d

    def Set_Seq_Lines(self, n, w):
        if self.dis_input:
            Goal_Sequence.Set_Seq_Lines(self, 0, 
                    IN_SEQ_WIDTH - EXPANSION_WIDTH)
            self.max_in_ring_buffer = 0
        else:
            Goal_Sequence.Set_Seq_Lines(self, n, w)
            

    def Update(self):
        self.ring_buffer.clear()
        self.dirty = True
        if self.dis_input:
            return 0

        if self.network == None:
            return 0

        in_list = []
        for in_list in self.network.Get_Stimulator_Sequence():
            for j in xrange(CLOCK_CYCLE_LENGTH - 1):
                self.ring_buffer.append([
                    (instate, STIM_COLOUR[ i % len(STIM_COLOUR) ])
                        for (i, instate) in enumerate(in_list) ])
            self.ring_buffer.append([ RB_CLOCK_EVENT
                        for s in self.network.Get_String_Data() ])

        return len(in_list)



class Sequence_Controller:
    def __init__(self, tree, logical = None, goalseq = None, 
                                outseq = None, inseq = None):
        assert isinstance(tree, Tree)
        assert (logical == None) or isinstance(logical, Logical)
        assert (goalseq == None) or isinstance(goalseq, Goal_Sequence)
        assert (outseq == None) or isinstance(outseq, Out_Sequence)
        assert (inseq == None) or isinstance(inseq, Goal_Sequence)

        self.tree = tree
        self.logical = logical
        self.goalseq = goalseq
        self.outseq = outseq
        self.inseq = inseq
        self.elements = [ tree, logical, goalseq, outseq, inseq ]
        self.cycle = CLOCK_CYCLE_LENGTH

    def Unset_Network(self):
        for e in self.elements:
            if e != None:
                e.Unset_Network()
        self.network = None

    def Set_Network_and_Strings(self, n, level_strings):
        for e in self.elements:
            if e != None:
                e.Set_Network(n)
        self.network = n
        self.network.Set_Strings(level_strings)
        sz = len(level_strings)
        if self.outseq != None:
            self.outseq.Set_Seq_Lines(len(level_strings), 
                        TREE_SIZE[ 0 ])
            self.outseq.Set_Label_List(
                        [ s.name for s in level_strings ])
        if self.goalseq != None:
            sz = max(sz, self.goalseq.Update())
            self.goalseq.Set_Label_List(
                        [ s.name for s in level_strings ])
        if self.inseq != None:
            sz = max(sz, self.inseq.Update())
            self.inseq.Set_Label_List(self.network.Get_Stimulator_Names())

        if self.inseq != None:
            self.inseq.Set_Seq_Lines(sz, IN_SEQ_WIDTH)

        if self.goalseq != None:
            if ((self.inseq != None)
            and (self.inseq.dis_input)):
                w = IN_SEQ_WIDTH + EXPANSION_WIDTH
            else:
                w = IN_SEQ_WIDTH
            self.goalseq.Set_Seq_Lines(sz, w)

    def Tick(self):
        if self.network == None:
            return False

        self.cycle += 1

        # Output to Tree and Logical 
        assert self.tree != None
        clock_event = False
        while self.cycle >= CLOCK_CYCLE_LENGTH:
            self.cycle -= CLOCK_CYCLE_LENGTH
            if self.network != None:
                self.network.Tick()

            clock_event = True

        if self.network.Get_Output_Change():
            if OVERPAINT_DEBUG:
                print '%1.2f Refresh Tree (GOC)' % (time.time() % 1000)
            self.tree.Refresh(False)

        # Output to Out_Sequence
        if clock_event:
            new_element = [ RB_CLOCK_EVENT
                            for s in self.network.Get_String_Data() ]
        elif self.network != None:
            new_element = [ (s.state, s.colour) 
                            for s in self.network.Get_String_Data() ]
        else:
            new_element = []

        if self.outseq != None:
            self.outseq.Push(new_element)
       
        # Output to Goal_Sequence
        if self.goalseq != None:
            gm = self.network.Get_Goal_Major()
            if gm == None:
                self.goalseq.Set_Thumb_Pos(-1)
            else:
                tp = (gm * CLOCK_CYCLE_LENGTH) + min(self.cycle,    
                        CLOCK_CYCLE_LENGTH - 1)
                self.goalseq.Set_Thumb_Pos(tp)

        if self.inseq != None:
            sm = self.network.Get_Stimulator_Major()
            assert sm != None
            tp = (sm * CLOCK_CYCLE_LENGTH) + min(self.cycle,    
                    CLOCK_CYCLE_LENGTH - 1)
            self.inseq.Set_Thumb_Pos(tp)

        return clock_event  

class Custom_Bitmap(wx.Window):
    def __init__(self, parent, *args1, **args2):
        wx.Window.__init__(self, parent, *args1, **args2)
        self.SetBackgroundColour(BACKGROUND_COLOUR)
        self.Bind(wx.EVT_PAINT, self.On_Paint)
        self.bmp = None

    def SetImage(self, name):
        self.bmp = wx.Bitmap(os.path.join(RO_DIR(), 
                    IMAGE_PATH, "%s.png" % name))

    def On_Paint(self, event):
        dc = wx.PaintDC(self)
        dc.DrawBitmap(self.bmp, 0, 0)

# Glue to wxglade-generated file

import wxglade_gui_output as wxg

wxg.Goal_Sequence = Goal_Sequence
wxg.Out_Sequence = Out_Sequence
wxg.In_Sequence = In_Sequence
wxg.Tree = Tree
wxg.Logical = Logical
wxg.CustomBitmap = Custom_Bitmap

from wxglade_gui_output import Top, Information, Complete_Level


